# IMPORT
import lightgbm as lgb
from sklearn.svm import SVC
from xgboost import XGBClassifier
from aiautomation.mlpackage.Entities import MultiClasModelEntity, MultiClassGridEntity
from catboost import CatBoostClassifier
from sklearn.tree import DecisionTreeClassifier
from sklearn.neighbors import KNeighborsClassifier
from aiautomation.mlpackage.PackageVariable import Variable
from sklearn.linear_model import LogisticRegression, RidgeClassifier
from sklearn.multiclass import OneVsOneClassifier
from sklearn.ensemble import RandomForestClassifier, GradientBoostingClassifier, AdaBoostClassifier


class Models:

    def __init__(self):
        self.alg_name = {}

    # define different models
    # LOGISTIC REGRESSION
    def log_reg(self):
        # define models and parameters
        # TODO max_iteration =1000
        logistic_reg = LogisticRegression(max_iter=1100)

        # multi class classification
        ovo = OneVsOneClassifier(logistic_reg)

        # Start the model to run
        self.alg_name[logistic_reg] = 'sklearn.linear_model.LogisticRegression'
        multi_class_grid_entity = self.check_and_get_grid_entity(Variable.typeLogReg)
        multi_class_model_entity = MultiClasModelEntity(ovo, multi_class_grid_entity.grid, Variable.typeLogReg,
                                                        multi_class_grid_entity.gene_grid, logistic_reg)
        return multi_class_model_entity

    # -----------------------------------------------------------------------------
    # RIDGE CLASSIFIER
    def rid_clas(self):
        # define models and parameters
        rid_clas = RidgeClassifier()

        # multi class classification
        ovo = OneVsOneClassifier(rid_clas)

        # Start the model to run
        self.alg_name[rid_clas] = 'sklearn.linear_model.RidgeClassifier'

        multi_class_grid_entity = self.check_and_get_grid_entity(Variable.typeRidge)
        multi_class_model_entity = MultiClasModelEntity(ovo, multi_class_grid_entity.grid, Variable.typeRidge,
                                                        multi_class_grid_entity.gene_grid, rid_clas)
        return multi_class_model_entity

    # -----------------------------------------------------------------------------
    # K NEAREST NEIGHBOUR
    def knn(self):
        # define models and parameters
        knn = KNeighborsClassifier()

        # multi class classification
        ovo = OneVsOneClassifier(knn)

        # Start the model to run
        self.alg_name[knn] = 'sklearn.neighbors.KNeighborsClassifier'

        multi_class_grid_entity = self.check_and_get_grid_entity(Variable.typeKnn)
        multi_class_model_entity = MultiClasModelEntity(ovo, multi_class_grid_entity.grid, Variable.typeKnn,
                                                        multi_class_grid_entity.gene_grid, knn)
        return multi_class_model_entity

    # -----------------------------------------------------------------------------
    # Support Vector Machine (SVM)
    def svm(self):
        # define models and parameters
        svc_classifier = SVC()

        # multi class classification
        ovo = OneVsOneClassifier(svc_classifier)

        # Start the model to run	
        self.alg_name[svc_classifier] = 'sklearn.svm.SVC'

        multi_class_grid_entity = self.check_and_get_grid_entity(Variable.typeSvc)
        multi_class_model_entity = MultiClasModelEntity(ovo, multi_class_grid_entity.grid, Variable.typeSvc,
                                                        multi_class_grid_entity.gene_grid, svc_classifier)
        return multi_class_model_entity

    # -----------------------------------------------------------------------------
    # DECISION TREE
    def des_tree(self):
        # define models and parameters
        des_tree_classifier = DecisionTreeClassifier()

        # multi class classification
        ovo = OneVsOneClassifier(des_tree_classifier)

        # Start the model to run
        self.alg_name[des_tree_classifier] = 'sklearn.tree.DecisionTreeClassifier'

        multi_class_grid_entity = self.check_and_get_grid_entity(Variable.typeDesTree)
        multi_class_model_entity = MultiClasModelEntity(ovo, multi_class_grid_entity.grid, Variable.typeDesTree,
                                                        multi_class_grid_entity.gene_grid,
                                                        des_tree_classifier)
        return multi_class_model_entity

    # -----------------------------------------------------------------------------
    # RANDOM FOREST
    def ran_for(self):
        # define models and parameters
        ran_for_classifier = RandomForestClassifier()

        # multi class classification
        ovo = OneVsOneClassifier(ran_for_classifier)

        # Start the model to runs
        self.alg_name[ran_for_classifier] = 'sklearn.ensemble.RandomForestClassifier'

        multi_class_grid_entity = self.check_and_get_grid_entity(Variable.typeRanFor)
        multi_class_model_entity = MultiClasModelEntity(ovo, multi_class_grid_entity.grid, Variable.typeRanFor,
                                                        multi_class_grid_entity.gene_grid, ran_for_classifier)
        return multi_class_model_entity

    '''#-----------------------------------------------------------------------------
    #GAUSSIAN NB
    def gaussianNB():
        start = time.time()
        print ("GAUSSIAN NAIVE BAYES START  ------------")

        ## define models and parameters
        gaussian_nb = GaussianNB()
        var_smoothing = [1e-9, 1e-8, 1e-7, 1e-6, 1e-5, 1e-4, 1e-3, 1e-2, 1e-1, 1e-0, 1]

        #define grid search
        grid = dict(var_smoothing = var_smoothing)

         #Start the model alg_name[gaussian_nb] = 'sklearn.naive_bayes.GaussianNB' t11 = threading.Thread(
         target=model_grid_fit,args=(gaussian_nb, X, Y,X_test,grid,"GAUSSIAN_NB_submission_grid")) t12 = 
         threading.Thread(target=model_bayes_fit,args = (gaussian_nb, X, Y,X_test,grid, 
         "GAUSSIAN_NB_submission_bayes")) t13 = threading.Thread(target=model_gene_fit,args = (gaussian_nb, X, Y,
         X_test,grid, "GAUSSIAN_NB_submission_gene")) t14 = threading.Thread(target=model_optuna_fit,
         args = (gaussian_nb, X, Y, X_test,grid,"GAUSSIAN_NB_submission_optuna")) t11.start() t12.start() t13.start() 
         t14.start() t11.join() t12.join() t13.join() t14.join() 

        end = time.time()	
        print ("Gaussian Naive bayes Finish. In time of seconds = ",int(end-start),"\n\n\n")


    #-----------------------------------------------------------------------------
    #MULTINOMIAL NB
    def multinomialNB():
        start = time.time()
        print ("Multinomial NAIVE BAYES START  ------------")

        ## define models and parameters
        multinomial_nb = MultinomialNB()
        alpha = [0.1,0.2,0.3,0.4,0.5,0.6,0.7,0.8,0.9,1.0]
        fit_prior = [False, True]

        # define grid search
        grid = dict(alpha = alpha, fit_prior = fit_prior)

         #Start the model alg_name[multinomial_nb] = 'sklearn.naive_bayes.MultinomialNB' t11 = threading.Thread(
         target=model_grid_fit,args=(multinomial_nb, X, Y,X_test,grid,"MULTINOMIAL_NB_submission_grid")) t12 = 
         threading.Thread(target=model_bayes_fit,args = (multinomial_nb, X, Y,X_test,grid,
         "MULTINOMIAL_NB_submission_bayes")) t13 = threading.Thread(target=model_gene_fit,args = (multinomial_nb, X, Y,
         X_test,grid,"MULTINOMIAL_NB_submission_gene")) t14 = threading.Thread(target=model_optuna_fit,
         args = (multinomial_nb, X, Y, X_test,grid,"MULTINOMIAL_NB_submission_optuna")) t11.start() t12.start() 
         t13.start() t14.start() t11.join() t12.join() t13.join() t14.join() 

        end = time.time()	
        print ("Multinomial Naive bayes Finish. In time of seconds = ",int(end-start),"\n\n\n")'''

    # -----------------------------------------------------------------------------
    # Stochastic Gradient Boosting
    def grad_boost(self):
        # define models and parameters
        grad_boost_class = GradientBoostingClassifier()

        # multi class classification
        ovo = OneVsOneClassifier(grad_boost_class)

        # Start the model
        self.alg_name[grad_boost_class] = 'sklearn.ensemble.GradientBoostingClassifier'

        multi_class_grid_entity = self.check_and_get_grid_entity(Variable.typeGradBoost)
        multi_class_model_entity = MultiClasModelEntity(ovo, multi_class_grid_entity.grid, Variable.typeGradBoost,
                                                        multi_class_grid_entity.gene_grid, grad_boost_class)
        return multi_class_model_entity

    # -----------------------------------------------------------------------------
    # Adaptive Boosting
    def ada_boost(self):
        """
        max_depth = [3,None]
        max_features = [1,2,3,4,5,6,7,8,9,10]
        min_sample_leaf = [1,2,3,4,5,6,7,8,9,10]
        criterion = ['gini','entropy']

        des_classifier = DecisionTreeClassifier(max_depth=max_depth,max_features=max_features,
        min_samples_leaf=min_sample_leaf,criterion=criterion) """
        ada_boost_class = AdaBoostClassifier()

        # multi class classification
        ovo = OneVsOneClassifier(ada_boost_class)

        # Start the model
        self.alg_name[ada_boost_class] = 'sklearn.ensemble.AdaBoostClassifier'

        multi_class_grid_entity = self.check_and_get_grid_entity(Variable.typeAdaBoost)
        multi_class_model_entity = MultiClasModelEntity(ovo, multi_class_grid_entity.grid, Variable.typeAdaBoost,
                                                        multi_class_grid_entity.gene_grid, ada_boost_class)
        return multi_class_model_entity

    # -----------------------------------------------------------------------------
    # XGB boost
    def xgb(self):
        # define models and parameters
        xgb_model = XGBClassifier()

        # multi class classification
        ovo = OneVsOneClassifier(xgb_model)

        # Start the model
        self.alg_name[xgb_model] = 'xgboost.XGBClassifier'

        multi_class_grid_entity = self.check_and_get_grid_entity(Variable.typeXGB)
        multi_class_model_entity = MultiClasModelEntity(ovo, multi_class_grid_entity.grid, Variable.typeXGB,
                                                        multi_class_grid_entity.gene_grid, xgb_model)
        return multi_class_model_entity

    # -----------------------------------------------------------------------------
    # Light GBM
    def light_gbm(self):
        # define models and parameters
        lightgbm = lgb.LGBMClassifier()

        # multi class classification
        ovo = OneVsOneClassifier(lightgbm)

        # Start the model
        self.alg_name[lightgbm] = 'lgb.LGBMClassifier'

        multi_class_grid_entity = self.check_and_get_grid_entity(Variable.typeLGBM)
        multi_class_model_entity = MultiClasModelEntity(ovo, multi_class_grid_entity.grid, Variable.typeLGBM,
                                                        multi_class_grid_entity.gene_grid, lightgbm)
        return multi_class_model_entity

    # -----------------------------------------------------------------------------
    # Cat Boost GBM
    def cat_boost_gbm(self):
        # define models and parameters
        catboost = CatBoostClassifier()

        # multi class classification
        ovo = OneVsOneClassifier(catboost)

        # Start the model
        # self.alg_name[catboost] = 'catboost.CatBoostClassifier'

        multi_class_grid_entity = self.check_and_get_grid_entity(Variable.typeCGBM)
        multi_class_model_entity = MultiClasModelEntity(ovo, multi_class_grid_entity.grid, Variable.typeCGBM,
                                                        multi_class_grid_entity.gene_grid, catboost)
        return multi_class_model_entity

    def get_all_models(self):
        return [self.rid_clas(), self.log_reg(), self.knn(), self.des_tree(),
                self.svm()]  # , self.ran_for(), self.ada_Boost(), self.grad_boost(), self.xgb(), self.light_gbm(),
        # self.cat_boost_gbm()]

    def get_algorithm_name(self):
        return self.alg_name

    def check_and_get_grid(self, model_type, search_type):
        multi_class_grid_entity = self.check_and_get_grid_entity(model_type)
        grid = multi_class_grid_entity.grid
        if search_type == Variable.typeGene:
            grid = multi_class_grid_entity.gene_grid
        return grid

    @staticmethod
    def check_and_get_grid_entity(model_type):
        multi_class_grid_entity = None
        if model_type == Variable.typeRidge:
            # Whole values
            alpha = [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0]

            # define grid search
            multi_class_grid_entity = MultiClassGridEntity(dict(estimator__alpha=alpha), dict(alpha=alpha))

        elif model_type == Variable.typeLogReg:
            # Whole values
            solvers = ['newton-cg', 'lbfgs', 'liblinear', 'sag', 'saga']
            penalty = ['l2']
            c_values = [100, 10, 1.0, 0.1, 0.01]

            # define grid search
            multi_class_grid_entity = MultiClassGridEntity(
                dict(estimator__solver=solvers, estimator__penalty=penalty, estimator__C=c_values)
                , dict(solver=solvers, penalty=penalty, C=c_values))

        elif model_type == Variable.typeKnn:
            # Whole values
            n_neighbors = list(range(1, 21, 2))
            weights = ['uniform', 'distance']
            metric = ['euclidean', 'manhattan', 'minkowski']

            # define grid search
            multi_class_grid_entity = MultiClassGridEntity(dict(estimator__n_neighbors=n_neighbors,
                                                                estimator__weights=weights, estimator__metric=metric),
                                                           dict(n_neighbors=n_neighbors, weights=weights,
                                                                metric=metric))

        elif model_type == Variable.typeSvc:
            # Whole values
            kernel = ['poly', 'rbf', 'sigmoid']
            c = [100, 10, 1.0, 0.1, 0.01]
            gamma = ['scale']

            # define grid search
            multi_class_grid_entity = MultiClassGridEntity(dict(estimator__kernel=kernel, estimator__C=c,
                                                                estimator__gamma=gamma), dict(kernel=kernel, C=c,
                                                                                              gamma=gamma))

        elif model_type == Variable.typeDesTree:
            # Whole values
            # Whole values
            max_depth = [3, None]
            max_features = list(range(1, 10, 1))
            min_sample_leaf = list(range(1, 10, 1))
            criterion = ['gini', 'entropy']

            # define grid search
            multi_class_grid_entity = MultiClassGridEntity(dict(estimator__max_depth=max_depth,
                                                                estimator__max_features=max_features,
                                                                estimator__min_samples_leaf=min_sample_leaf,
                                                                estimator__criterion=criterion),
                                                           dict(max_depth=max_depth, max_features=max_features,
                                                                min_samples_leaf=min_sample_leaf,
                                                                criterion=criterion))

        elif model_type == Variable.typeRanFor:
            # Whole values
            n_estimators = list(range(100, 1001, 100))
            max_features = ['sqrt', 'log2']

            # define grid search
            multi_class_grid_entity = MultiClassGridEntity(
                dict(estimator__n_estimators=n_estimators, estimator__max_features=max_features),
                dict(n_estimators=n_estimators, max_features=max_features))

        elif model_type == Variable.typeGradBoost:
            # Whole values
            n_estimators = list(range(100, 1000, 100))
            learning_rate = [0.001, 0.01, 0.1]
            subsample = [0.5, 0.7, 1.0]
            max_depth = [3, 7, 9]

            # define grid search
            multi_class_grid_entity = MultiClassGridEntity(
                dict(estimator__learning_rate=learning_rate, estimator__n_estimators=n_estimators,
                     estimator__subsample=subsample, estimator__max_depth=max_depth),
                dict(learning_rate=learning_rate, n_estimators=n_estimators, subsample=subsample,
                     max_depth=max_depth))

        elif model_type == Variable.typeAdaBoost:
            n_estimators = list(range(50, 1001, 50))
            learning_rate = [0.001, 0.01, 0.1]
            algorithm = ['SAMME', 'SAMME.R']

            # define grid search
            multi_class_grid_entity = MultiClassGridEntity(
                dict(estimator__learning_rate=learning_rate, estimator__n_estimators=n_estimators,
                     estimator__algorithm=algorithm),
                dict(learning_rate=learning_rate, n_estimators=n_estimators, algorithm=algorithm))

        elif model_type == Variable.typeXGB:
            # Whole values
            max_depth = list(range(3, 10, 2))
            min_child_weight = list(range(1, 6, 2))
            gamma = [float(i / 10.0) for i in range(0, 5)]
            n_estimators = list(range(100, 1000, 100))
            learning_rate = [0.001, 0.01, 0.1]
            subsample = [i / 10.0 for i in range(6, 10)],
            col_sample_by_tree = [i / 10.0 for i in range(6, 10)]
            reg_alpha = [1e-5, 1e-2, 0.1, 1, 100, 0, 0.001, 0.005, 0.01, 0.05]

            # define grid search
            multi_class_grid_entity = MultiClassGridEntity(
                dict(estimator__learning_rate=learning_rate, estimator__n_estimators=n_estimators,
                     estimator__subsample=subsample, estimator__max_depth=max_depth,
                     estimator__min_child_weight=min_child_weight, estimator__gamma=gamma,
                     estimator__colsample_bytree=col_sample_by_tree, estimator__reg_alpha=reg_alpha),
                dict(learning_rate=learning_rate, n_estimators=n_estimators, subsample=subsample,
                     max_depth=max_depth, min_child_weight=min_child_weight, gamma=gamma,
                     colsample_bytree=col_sample_by_tree, reg_alpha=reg_alpha))

        elif model_type == Variable.typeLGBM:
            # Whole values
            num_leaves = list(range(50, 1000, 50))
            learning_rate = [0.001, 0.01, 0.1]
            max_depth = list(range(3, 10, 2))
            max_bin = list(range(100, 1000, 100))
            objective = 'binary'

            # define grid search
            multi_class_grid_entity = MultiClassGridEntity(
                dict(estimator__learning_rate=learning_rate, estimator__num_leaves=num_leaves,
                     estimator__max_bin=max_bin, estimator__max_depth=max_depth, estimator__objective=objective),
                dict(learning_rate=learning_rate, num_leaves=num_leaves, max_bin=max_bin, max_depth=max_depth,
                     objective=objective))

        elif model_type == Variable.typeCGBM:
            # Whole values
            depth = [3, 1, 2, 6, 4, 5, 7, 8, 9, 10]
            iterations = [250, 100, 500, 1000]
            learning_rate = [0.03, 0.001, 0.01, 0.1, 0.2, 0.3]
            l2_leaf_reg = [3, 1, 5, 10, 100]
            border_count = [32, 5, 10, 20, 50, 100, 200],
            ctr_border_count = [50, 5, 10, 20, 100, 200],
            thread_count = 4

            # define grid search
            multi_class_grid_entity = MultiClassGridEntity(
                dict(estimator__learning_rate=learning_rate, estimator__depth=depth,
                     estimator__iterations=iterations,
                     estimator__l2_leaf_reg=l2_leaf_reg, estimator__border_count=border_count,
                     estimator__ctr_border_count=ctr_border_count, estimator__thread_count=thread_count),
                dict(learning_rate=learning_rate, depth=depth, iterations=iterations, l2_leaf_reg=l2_leaf_reg,
                     border_count=border_count, ctr_border_count=ctr_border_count, thread_count=thread_count))

        return multi_class_grid_entity



# IMPORT
import matplotlib
import numpy as np
import optuna
import wandb
from aiautomation.mlpackage.Entities import AccuracyEntity
from aiautomation.mlpackage.PackageVariable import Variable
from hyperopt import hp, fmin, tpe, STATUS_OK, Trials
from sklearn.model_selection import GridSearchCV, cross_val_score
from sklearn.utils.extmath import softmax
from sklearn_genetic import GASearchCV
from sklearn_genetic.space import Categorical, Continuous
from tpot import TPOTClassifier, TPOTRegressor
from yellowbrick.model_selection import LearningCurve, ValidationCurve


class HyperParameterTuning:

    def __init__(self, hp_entity):
        self.hPEntity = hp_entity
        print("X_TRAIN SHAPE = ", hp_entity.x_train.shape)
        print("Y_TRAIN SHAPE = ", hp_entity.y_train.shape)
        self.best_score = 0
        self.best_val_score = 0
        self.io = None
        self.actualFolderName = ""
        # wandb.init(project="visualize-sklearn")

    def set_actual_folder_name(self, folder_name):
        self.actualFolderName = folder_name

    def init_wand_b(self, name, custom_dir):
        wandb.init(project=self.actualFolderName, name=name, save_code=True, dir=custom_dir, group=name, mode="offline")

    def update_io(self, io):
        self.io = io

    # search space for bayesian
    @staticmethod
    def search_space(grid):
        space = {}

        for i in grid.keys():
            param = hp.choice(str(i), grid[i])
            space[i] = param
        return space

        # search space for bayesian

    @staticmethod
    def ga_search_param(grid):
        space = {}

        for i in grid.keys():
            if isinstance(grid[i], list) and isinstance(grid[i][0], float):
                param = Continuous(grid[i][0], grid[i][-1])
            else:
                param = Categorical(grid[i])
            space[i] = param
        return space

    def set_scores(self, score, val_score):
        self.best_score = score
        self.best_val_score = val_score

    @staticmethod
    def get_bayes_param(best, grid):
        new_grid = {}
        for i in best:
            new_grid[i] = grid[i][best[i]]
        return new_grid

    def set_scoring(self, scoring):
        self.hPEntity.scoring = scoring

    def get_all_scoring(self):
        return self.hPEntity.scoring

    ###################################################################################################################
    # Fit the model
    # GRID SEARCH
    def model_grid_fit(self, model_entity, file_name, folder_path):
        grid_search = GridSearchCV(estimator=model_entity.alg, param_grid=model_entity.grid, n_jobs=1,
                                   cv=self.hPEntity.cv, scoring=self.hPEntity.scoring, error_score=0)

        grid_search.fit(self.hPEntity.x_train, self.hPEntity.y_train)

        score = 0
        val_score = 0
        if self.hPEntity.model_type != Variable.typeSegmentation:
            score = grid_search.best_score_
            val_score = grid_search.score(self.hPEntity.x_val, self.hPEntity.y_val)

        self.check_and_save_model(score, val_score, file_name, folder_path, grid_search)

        return AccuracyEntity(file_name, score, val_score, str(grid_search.best_params_))

    # BAYESIAN OPTIMIZATION # https://www.kaggle.com/code/prashant111/bayesian-optimization-using-hyperopt/notebook
    def model_bayes_fit(self, model_entity, file_name, folder_path):

        x_train = self.hPEntity.x_train
        y_train = self.hPEntity.y_train
        model = model_entity.alg
        grid = model_entity.grid

        # Objective
        def objective():
            temp_score = cross_val_score(model, x_train, y_train, cv=self.hPEntity.cv,
                                         scoring=self.hPEntity.scoring).mean()
            # We aim to maximize accuracy, therefore we return it as a negative value
            return {'loss': -temp_score, 'status': STATUS_OK}

        trials = Trials()
        best = fmin(fn=objective,
                    algo=tpe.suggest,
                    max_evals=80,
                    space=self.search_space(grid),
                    trials=trials)

        # setting the params for model
        param = self.get_bayes_param(best, grid)
        alg = model.set_params(**param)
        alg.fit(x_train, y_train)

        score = 0
        val_score = 0
        if self.hPEntity.model_type != Variable.typeSegmentation:
            score = trials.best_trial['result']['loss']
            val_score = alg.score(self.hPEntity.x_val, self.hPEntity.y_val)

        self.check_and_save_model(score, val_score, file_name, folder_path, alg)

        return AccuracyEntity(file_name, score, val_score, str(best))

    # GENETIC ALGORITHM
    def model_gene_fit(self, model_entity, file_name, folder_path):

        """accuracy, average_precision, roc_auc, recall"""
        x_train = self.hPEntity.x_train
        y_train = self.hPEntity.y_train
        cv = self.hPEntity.cv
        scoring = self.hPEntity.scoring
        model_type = self.hPEntity.model_type

        if model_type == Variable.typeMultiClass:
            model = model_entity.gene_alg
            grid = model_entity.gene_grid
        else:
            model = model_entity.alg
            grid = model_entity.grid

        if model_type == Variable.typeRegress:
            tpot = TPOTRegressor(generations=5, population_size=24, offspring_size=12, verbosity=2, early_stop=12,
                                 config_dict={self.hPEntity.alg_name[model]: grid}, cv=cv, scoring=scoring)
        else:
            tpot = TPOTClassifier(generations=5, population_size=24, offspring_size=12, verbosity=2, early_stop=12,
                                  config_dict={self.hPEntity.alg_name[model]: grid}, cv=cv, scoring=scoring)

        model = tpot.fit(x_train, y_train)

        # TODO REMOVE THE BELOW COMMENTED CODE AND TRY TO GET GENE FILE PATH LOCATION ACCORDING TO RUN ALL FILE
        self.export_gene(model, file_name)

        score = 0
        val_score = 0
        if model_type != Variable.typeSegmentation:
            score = model.score(x_train, y_train)
            val_score = model.score(self.hPEntity.x_val, self.hPEntity.y_val)

        self.check_and_save_model(score, val_score, file_name, folder_path, model)

        return AccuracyEntity(file_name, score, val_score, "example")

    # OPTUNA ALGORITHM
    def model_optuna_fit(self, model_entity, file_name, folder_path):

        model = model_entity.alg
        grid = model_entity.grid
        x_train = self.hPEntity.x_train
        y_train = self.hPEntity.y_train

        def objective(temp_trial):
            space = {}
            for i in grid.keys():
                space[i] = temp_trial.suggest_categorical(str(i), grid[i])
            model.set_params(**space)
            temp_score = cross_val_score(model, x_train, y_train, cv=self.hPEntity.cv,
                                         scoring=self.hPEntity.scoring).mean()
            return temp_score

        study = optuna.create_study(direction='maximize')
        study.optimize(objective, n_trials=100)

        trial = study.best_trial

        # setting the params for model
        alg = model.set_params(**trial.params)
        alg.fit(x_train, y_train)

        score = 0
        val_score = 0
        if self.hPEntity.model_type != Variable.typeSegmentation:
            score = trial.value
            val_score = alg.score(self.hPEntity.x_val, self.hPEntity.y_val)

        self.check_and_save_model(score, val_score, file_name, folder_path, alg)

        return AccuracyEntity(file_name, score, val_score, str(trial.params))

    # GENETIC ALGORITHM
    def model_sklearn_genetic_fit(self, model_entity, file_name, folder_path):
        ga_search = GASearchCV(estimator=model_entity.alg, param_grid=self.ga_search_param(model_entity.grid), n_jobs=1,
                               cv=self.hPEntity.cv, scoring=self.hPEntity.scoring)

        ga_search.fit(self.hPEntity.x_train, self.hPEntity.y_train)
        model = ga_search.best_estimator_

        score = 0
        val_score = 0
        if self.hPEntity.model_type != Variable.typeSegmentation:
            score = model.score(self.hPEntity.x_train, self.hPEntity.y_train)
            val_score = model.score(self.hPEntity.x_val, self.hPEntity.y_val)

        self.check_and_save_model(score, val_score, file_name, folder_path, model)

        return AccuracyEntity(file_name, score, val_score, str(ga_search.best_params_))

    def check_and_save_model(self, score, val_score, model_name, folder_path, model):
        self.save_all_model(model_name, folder_path, model)
        if self.can_save_model(score, val_score):
            self.save_best_model(model_name, model)
            self.best_score = score
            self.best_val_score = val_score

    def export_gene(self, tpot, file_name):
        self.io.export_gene(tpot, file_name)

    def save_best_model(self, model_name, model):
        self.io.save_best_model(model_name, model)

    def save_all_model(self, model_name, folder_path, model):
        self.io.save_all_model(model_name, folder_path, model)

    def save_all_visualizer(self, model_name, folder_path, model):
        self.io.save_all_visualizer(model_name, folder_path, model)

    def can_save_model(self, score, val_score):
        if score > 0 and val_score > 0:
            if self.get_check_score(self.best_val_score, val_score):
                return True
            if self.best_val_score == val_score:
                if self.get_check_score(self.best_score, score) or self.best_score == score:
                    return True

    def get_check_score(self, best_score, score):
        if self.hPEntity.model_type == Variable.typeRegress:
            return best_score > score
        else:
            return best_score < score

    def start_visualization(self, model, cv, scoring, x_train, y_train, x_val, y_val, grid, model_name, folder_path):

        matplotlib.use('Agg')
        self.start_learning_curve(model, cv, scoring, x_train, y_train, model_name, folder_path)
        print("\n")
        self.start_validation_curve(model, cv, scoring, x_val, y_val, grid, model_name, folder_path)

    def start_learning_curve(self, model, cv, scoring, x, y, model_name, folder_path):
        visualizer = LearningCurve(model, cv=cv, scoring=scoring, n_jobs=-1)
        visualizer.fit(x, y)  # Fit the data to the visualizer
        model_name = model_name + Variable.fileSeparator + Variable.learningCurve
        self.save_all_visualizer(model_name, folder_path, visualizer)
        # visualizer.savefig("mygraph.png")

    def start_validation_curve(self, model, cv, scoring, x, y, grid, model_name, folder_path):
        for key, value in grid.items():
            visualizer = ValidationCurve(model, cv=cv, scoring=scoring, n_jobs=-1, param_name=key, param_range=value)
            visualizer.fit(x, y)  # Fit the data to the visualizer
            model_name = model_name + Variable.fileSeparator + key + Variable.fileSeparator + Variable.validationCurve
            self.save_all_visualizer(model_name, folder_path, visualizer)
        # visualizer.show()

    def start_wandb(self, model, x_train, y_train, x_val, y_val, model_name, model_type):
        wandb_dir = self.io.storeDataDirName + Variable.locationSeparator + Variable.dataFolderName
        self.init_wand_b(model_name, wandb_dir)
        wandb.sklearn.plot_learning_curve(model, x_train, y_train)

        if model_type == Variable.typeSegmentation:
            # All clustering plots
            wandb.sklearn.plot_clusterer(model, x_train, cluster_labels, labels=None, model_name=model_name)

        elif model_type == Variable.typeRegress:
            # All regression plots
            wandb.sklearn.plot_regressor(model, x_train, x_val, y_train, y_val, model_name=model_name)

        else:
            # Visualize all classifier plots
            wandb.sklearn.plot_classifier(model, x_train, x_val, y_train, y_val, y_pred=model.predict(x_val),
                                          y_probas=self.get_predict_proba(model, x_val), labels=self.hPEntity.labels,
                                          model_name=model_name, feature_names=None)

        wandb.finish()

    @staticmethod
    def get_predict_proba(model, x_val):
        try:
            return model.predict_proba(x_val)
        except:
            d = model.decision_function(x_val)
            d_2d = np.c_[-d, d]
            return softmax(d_2d)

    def run_all_visualization(self, model_visualize_entity, folder_path):
        self.run_all_visualization(self.hPEntity.cv, self.hPEntity.x_train,
                                   self.hPEntity.y_train, self.hPEntity.x_val, self.hPEntity.y_val,
                                   model_visualize_entity.grid, model_visualize_entity.file_name,
                                   self.hPEntity.model_type, folder_path)

    def run_all_visualization(self, model, cv, scoring, x_train, y_train, x_val, y_val, grid, model_name, model_type,
                              folder_path):
        print("\nstart Visualization\n")
        self.start_visualization(model, cv, scoring, x_train, y_train, x_val, y_val, grid, model_name, folder_path)
        self.start_wandb(model, x_train, y_train, x_val, y_val, model_name, model_type)
        print("\nfinished Visualization\n")



# IMPORT
from sklearn.cluster import AgglomerativeClustering
from sklearn.cluster import Birch
from sklearn.cluster import KMeans
from sklearn.cluster import MiniBatchKMeans
from sklearn.cluster import SpectralClustering
from sklearn.mixture import GaussianMixture
from aiautomation.mlpackage.Entities import ClasRegModelEntity
from aiautomation.mlpackage.PackageVariable import Variable


class Models:

    def __init__(self, cluster):
        self.alg_name = {}
        self.clusters = cluster

    # define different models
    def agglo_cluster(self):
        # define models and parameters
        model = AgglomerativeClustering(n_clusters=self.clusters)

        clas_reg_model_entity = ClasRegModelEntity(model, self.check_and_get_grid(Variable.typeAglo),
                                                    Variable.typeAglo)
        self.alg_name[model] = 'sklearn.cluster.AgglomerativeClustering'

        return clas_reg_model_entity

    # Balanced Iterative Reducing and Clustering
    def birch(self):
        # define models and parameters
        model = Birch(n_clusters=self.clusters)

        clas_reg_model_entity = ClasRegModelEntity(model, self.check_and_get_grid(Variable.typeBirch),
                                                    Variable.typeBirch)
        self.alg_name[model] = 'sklearn.cluster.Birch'

        return clas_reg_model_entity

    # Kmeans
    def kmeans(self):
        # define the model
        model = KMeans(n_clusters=self.clusters, n_jobs=-1)

        clas_reg_model_entity = ClasRegModelEntity(model, self.check_and_get_grid(Variable.typeKmeans),
                                                    Variable.typeKmeans)
        self.alg_name[model] = 'sklearn.cluster.KMeans'

        return clas_reg_model_entity

    # Mini-Batch K-Means
    def min_batch_k(self):
        # define the model
        model = MiniBatchKMeans(n_clusters=self.clusters)

        clas_reg_model_entity = ClasRegModelEntity(model, self.check_and_get_grid(Variable.typeMiniBatch),
                                                    Variable.typeMiniBatch)
        self.alg_name[model] = 'sklearn.cluster.MiniBatchKMeans'

        return clas_reg_model_entity

    # Spectral Clustering
    def spec_cluster(self):
        # define the model
        model = SpectralClustering(n_clusters=self.clusters, n_jobs=-1)

        clas_reg_model_entity = ClasRegModelEntity(model, self.check_and_get_grid(Variable.typeSpecCluster),
                                                    Variable.typeSpecCluster)
        self.alg_name[model] = 'sklearn.cluster.SpectralClustering'

        return clas_reg_model_entity

    # Gaussian Mixture Model
    def gaussian_mix(self):
        # define the model
        model = GaussianMixture(n_components=self.clusters)

        clas_reg_model_entity = ClasRegModelEntity(model, self.check_and_get_grid(Variable.typeGaussianMix),
                                                    Variable.typeGaussianMix)
        self.alg_name[model] = 'sklearn.mixture.GaussianMixture'

        return clas_reg_model_entity

    @staticmethod
    def check_and_get_grid(model_type):
        grid = dict()
        if model_type == Variable.typeAglo:
            # Whole values
            affinity = ["euclidean", "l1", "l2", "manhattan", "cosine", "precomputed"]
            compute = ["auto", True, False]
            linkage = ["ward", "complete", "average", "single"]
            threshold = [0.0, 1e-0, 1e-1, 1e-2, 1e-3, 1e-4, 1e-5, 1e-6, 1e-7, 1e-8, 1e-9, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7,
                         0.8, 0.9]
            distance = [True, False]

            grid = dict(affinity=affinity, compute_full_tree=compute, linkage=linkage, compute_distances=distance,
                        distance_threshold=threshold)

        elif model_type == Variable.typeBirch:
            threshold = [0.0, 1e-0, 1e-1, 1e-2, 1e-3, 1e-4, 1e-5, 1e-6, 1e-7, 1e-8, 1e-9, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7,
                         0.8, 0.9]
            branching_factor = list(range(5, 100, 5))
            labels = [True, False]

            grid = dict(threshold=threshold, branching_factor=branching_factor, compute_labels=labels)

        elif model_type == Variable.typeKmeans:
            n_init = list(range(0, 100, 5))
            # max_iter = list(range(0, 1000, 50))
            tol = [0.0, 1e-0, 1e-1, 1e-2, 1e-3, 1e-4, 1e-5, 1e-6, 1e-7, 1e-8, 1e-9, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8,
                   0.9]
            distance = ["auto", True, False]
            algorithm = ["auto", "full", "elkan"]

            grid = dict(n_init=n_init, tol=tol, algorithm=algorithm, precompute_distances=distance)

        elif model_type == Variable.typeMiniBatch:
            max_iter = list(range(0, 1000, 50))
            batch_size = list(range(0, 1000, 50))
            labels = [True, False]
            tol = [0.0, 1e-0, 1e-1, 1e-2, 1e-3, 1e-4, 1e-5, 1e-6, 1e-7, 1e-8, 1e-9, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8,
                   0.9]
            improvement = list(range(0, 100, 5))
            reassignment = [0.0, 1e-0, 1e-1, 1e-2, 1e-3, 1e-4, 1e-5, 1e-6, 1e-7, 1e-8, 1e-9, 0.2, 0.3, 0.4, 0.5, 0.6,
                            0.7, 0.8, 0.9]

            grid = dict(max_iter=max_iter, batch_size=batch_size, tol=tol, compute_labels=labels,
                        max_no_improvement=improvement, reassignment_ratio=reassignment)

        elif model_type == Variable.typeSpecCluster:
            eigen_solver = ["arpack", "lobpcg", "amg", None]
            n_init = list(range(0, 100, 5))
            gamma = [0.0, 1e-0, 1e-1, 1e-2, 1e-3, 1e-4, 1e-5, 1e-6, 1e-7, 1e-8, 1e-9, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8,
                     0.9, 1.0]
            n_neighbors = list(range(0, 100, 5))
            eigen_tol = [0.0, 1e-0, 1e-1, 1e-2, 1e-3, 1e-4, 1e-5, 1e-6, 1e-7, 1e-8, 1e-9, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7,
                         0.8, 0.9, 1.0]
            labels = ["kmeans", "discretize"]
            affinity = ["nearest_neighbors", "rbf", "precomputed", "precomputed_nearest_neighbors"]

            grid = dict(eigen_solver=eigen_solver, n_init=n_init, gamma=gamma, affinity=affinity,
                        n_neighbors=n_neighbors, eigen_tol=eigen_tol, assign_labels=labels)

        elif model_type == Variable.typeGaussianMix:
            covariance_type = ["full", "tied", "diag", "spherical"]
            tol = [0.0, 1e-0, 1e-1, 1e-2, 1e-3, 1e-4, 1e-5, 1e-6, 1e-7, 1e-8, 1e-9, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8,
                   0.9]
            reg_covar = [0.0, 1e-0, 1e-1, 1e-2, 1e-3, 1e-4, 1e-5, 1e-6, 1e-7, 1e-8, 1e-9, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7,
                         0.8, 0.9]
            # max_iter = list(range(0, 1000, 50))
            n_init = list(range(0, 100, 5))
            init_params = ["kmeans", "random"]
            warm_start = [True, False]

            grid = dict(covariance_type=covariance_type, tol=tol, reg_covar=reg_covar, n_init=n_init,
                        init_params=init_params, warm_start=warm_start)

        return grid

    def get_all_models(self):
        return [self.birch(), self.min_batch_k(), self.gaussian_mix(), self.spec_cluster(), self.kmeans(),
                self.agglo_cluster()]

    def get_algorithm_name(self):
        # alg_name = self.alg_name
        # del self.alg_name
        return self.alg_name




from setuptools import setup

setup(
    name='ai_automation',
    version='3.1',
    license='MIT',
    author="SheikhChili",
    author_email='sparrow.shreyas@gmail.com',
    packages=['aiautomation', 'aiautomation/mlpackage/classification', 'aiautomation/mlpackage',
              'aiautomation/mlpackage/multiclassification', 'aiautomation/mlpackage/regression',
              'aiautomation/mlpackage/segmentation'],
    # package_dir={'': 'src'},
    url='https://github.com/SheikhChili/aipackage',
    keywords='AI Automation',
    install_requires=[
        'scikit-learn', 'xgboost', 'catboost', 'lightgbm', 'numpy', 'pandas==1.4.2', 'tensorflow', 'imblearn', 'optuna',
        'tpot', 'hyperopt', 'yellowbrick', 'librosa', 'keras_tuner', 'spacy', 'unidecode', 'bs4', 'dtale',
        'lime', 'matplotlib', 'shap', 'sweetviz', 'aix360', 'explainerdashboard', 'pandas_profiling',
        'pandas_visual_analysis', 'markupsafe==2.0.1'
    ],

)
